@@ -1,6 +1,6 @@
 // -*- coding: utf-8 -*-
 //
-// Copyright (C) 2024 Michael Büsch <m@bues.ch>
+// Copyright (C) 2024 - 2026 Michael Büsch <m@bues.ch>
 //
 // Licensed under the Apache License version 2.0
 // or the MIT license, at your option.
@@ -169,6 +169,103 @@ impl Resource {
             users.contains(&id)
         }
     }
+
+    pub fn timeout(&self) -> Option<Duration> {
+        match self {
+            Self::Port { timeout, .. } => *timeout,
+            Self::Jump { timeout, .. } => *timeout,
+        }
+    }
+}
+
+impl std::fmt::Display for Resource {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
+        fn write_users(
+            f: &mut std::fmt::Formatter<'_>,
+            users: &[UserId],
+        ) -> Result<(), std::fmt::Error> {
+            if !users.is_empty() {
+                write!(f, "  Users: ")?;
+                for (i, user) in users.iter().enumerate() {
+                    if i != 0 {
+                        write!(f, ", ")?;
+                    }
+                    write!(f, "{user}")?;
+                }
+                writeln!(f)?;
+            }
+            Ok(())
+        }
+
+        fn write_timeout(
+            f: &mut std::fmt::Formatter<'_>,
+            timeout: &Option<Duration>,
+        ) -> Result<(), std::fmt::Error> {
+            if let Some(timeout) = timeout {
+                writeln!(f, "  timeout: {} s", timeout.as_secs())?;
+            }
+            Ok(())
+        }
+
+        fn write_chain(
+            f: &mut std::fmt::Formatter<'_>,
+            chain: &Option<String>,
+            match_saddr: bool,
+            name: &str,
+        ) -> Result<(), std::fmt::Error> {
+            if let Some(chain) = chain {
+                let match_saddr = if match_saddr { " match-saddr" } else { "" };
+                writeln!(f, "  {name}-chain: {chain}{match_saddr}")?;
+            }
+            Ok(())
+        }
+
+        match self {
+            Self::Port {
+                id,
+                port,
+                tcp,
+                udp,
+                timeout,
+                users,
+            } => {
+                let tcpudp = if *tcp && *udp {
+                    "TCP/UDP"
+                } else if *tcp {
+                    "TCP"
+                } else if *udp {
+                    "UDP"
+                } else {
+                    ""
+                };
+                writeln!(f, "Port resource:")?;
+                writeln!(f, "  id: {id}")?;
+                writeln!(f, "  port: {port} {tcpudp}")?;
+                write_timeout(f, timeout)?;
+                write_users(f, users)?;
+            }
+            Self::Jump {
+                id,
+                input,
+                input_match_saddr,
+                forward,
+                forward_match_saddr,
+                output,
+                output_match_saddr,
+                timeout,
+                users,
+            } => {
+                writeln!(f, "Jump resource:")?;
+                writeln!(f, "  id: {id}")?;
+                write_chain(f, input, *input_match_saddr, "input")?;
+                write_chain(f, forward, *forward_match_saddr, "forward")?;
+                write_chain(f, output, *output_match_saddr, "output")?;
+                write_timeout(f, timeout)?;
+                write_users(f, users)?;
+            }
+        }
+        Ok(())
+    }
 }
 
 /// Error reporting policy.
@@ -845,11 +942,25 @@ impl Config {
         self.seccomp
     }
 
+    /// Get a list of all configured users.
+    pub fn users(&self) -> Vec<UserId> {
+        let mut users: Vec<UserId> = self.keys.keys().cloned().collect();
+        users.sort();
+        users
+    }
+
     /// Get a key value by key identifier from the `[KEYS]` section.
     pub fn key(&self, id: UserId) -> Option<&Key> {
         self.keys.get(&id)
     }
 
+    /// Get a list of all configured resources.
+    pub fn resources(&self) -> Vec<Resource> {
+        let mut resources: Vec<Resource> = self.resources.values().cloned().collect();
+        resources.sort_by_key(|r| r.id());
+        resources
+    }
+
     /// Get a resource value by resource identifier from the `[RESOURCES]` section.
     pub fn resource(&self, id: ResourceId) -> Option<&Resource> {
         self.resources.get(&id)
