@@ -1,6 +1,6 @@
 // -*- coding: utf-8 -*-
 //
-// Copyright (C) 2024 Michael Büsch <m@bues.ch>
+// Copyright (C) 2024 - 2026 Michael Büsch <m@bues.ch>
 //
 // Licensed under the Apache License version 2.0
 // or the MIT license, at your option.
@@ -9,10 +9,9 @@
 use anyhow as ah;
 use letmein_conf::Config;
 use letmein_proto::{secure_random, Key, UserId};
-use std::sync::Arc;
 
 /// Generate a new truly random and secure user key.
-pub async fn run_genkey(conf: Arc<Config>, user: Option<UserId>) -> ah::Result<()> {
+pub async fn run_genkey(conf: &Config, user: Option<UserId>) -> ah::Result<()> {
     let user = user.unwrap_or_else(|| conf.default_user());
     let key: Key = secure_random();
     let key: Vec<String> = key.iter().map(|b| format!("{b:02X}")).collect();
