@@ -1,6 +1,6 @@
 // -*- coding: utf-8 -*-
 //
-// Copyright (C) 2024 Michael Büsch <m@bues.ch>
+// Copyright (C) 2024 - 2026 Michael Büsch <m@bues.ch>
 //
 // Licensed under the Apache License version 2.0
 // or the MIT license, at your option.
@@ -8,8 +8,8 @@
 
 pub mod nftables;
 
-use anyhow as ah;
-use letmein_conf::Config;
+use anyhow::{self as ah, format_err as err};
+use letmein_conf::{Config, Resource};
 use std::{
     collections::HashMap,
     net::IpAddr,
@@ -55,6 +55,41 @@ impl std::fmt::Display for LeasePort {
     }
 }
 
+impl TryFrom<Resource> for LeasePort {
+    type Error = ah::Error;
+
+    fn try_from(res: Resource) -> ah::Result<Self> {
+        if let Resource::Port {
+            id: _,
+            port,
+            tcp,
+            udp,
+            timeout: _,
+            users: _,
+        } = res
+        {
+            match (tcp, udp) {
+                (true, false) => Ok(Self::Tcp(port)),
+                (false, true) => Ok(Self::Udp(port)),
+                (true, true) => Ok(Self::TcpUdp(port)),
+                (false, false) => Err(err!("LeasePort: Invalid port info in config.")),
+            }
+        } else {
+            Err(err!("LeasePort: Resource is not a Port resource"))
+        }
+    }
+}
+
+impl LeasePort {
+    pub fn port(&self) -> u16 {
+        match self {
+            Self::Tcp(port) => *port,
+            Self::Udp(port) => *port,
+            Self::TcpUdp(port) => *port,
+        }
+    }
+}
+
 /// TCP or UDP port number.
 #[derive(Clone, Copy, PartialEq, Eq)]
 enum SingleLeasePort {
@@ -281,11 +316,44 @@ impl std::fmt::Display for FirewallJumpTargets {
     }
 }
 
+impl TryFrom<Resource> for FirewallJumpTargets {
+    type Error = ah::Error;
+
+    fn try_from(res: Resource) -> ah::Result<Self> {
+        if let Resource::Jump {
+            id: _,
+            input,
+            input_match_saddr,
+            forward,
+            forward_match_saddr,
+            output,
+            output_match_saddr,
+            timeout: _,
+            users: _,
+        } = res
+        {
+            Ok(Self {
+                input,
+                input_match_saddr,
+                forward,
+                forward_match_saddr,
+                output,
+                output_match_saddr,
+            })
+        } else {
+            Err(err!("FirewallJumpTargets: Resource is not a Jump resource"))
+        }
+    }
+}
+
 /// Firewall action operations.
 pub trait FirewallAction {
     /// Add a rule to open the specified `port` for the specified `remote_addr`.
+    ///
     /// This operation shall handle the case where there already is such
     /// a rule present gracefully.
+    ///
+    /// Warning: This function is *not* async cancel safe.
     async fn open_port(
         &self,
         conf: &Config,
@@ -294,9 +362,25 @@ pub trait FirewallAction {
         timeout: Option<Duration>,
     ) -> ah::Result<()>;
 
+    /// Revoke/remove a port rule.
+    ///
+    /// This operation shall handle the case where there is no such
+    /// rule present gracefully.
+    ///
+    /// Warning: This function is *not* async cancel safe.
+    async fn revoke_port(
+        &self,
+        conf: &Config,
+        remote_addr: IpAddr,
+        port: LeasePort,
+    ) -> ah::Result<()>;
+
     /// Add jump rules to jump to the specified processing rule chains.
+    ///
     /// This operation shall handle the case where there already is such
     /// a rule present gracefully.
+    ///
+    /// Warning: This function is *not* async cancel safe.
     async fn add_jump(
         &self,
         conf: &Config,
@@ -304,6 +388,19 @@ pub trait FirewallAction {
         targets: &FirewallJumpTargets,
         timeout: Option<Duration>,
     ) -> ah::Result<()>;
+
+    /// Revoke/remove a jump rule.
+    ///
+    /// This operation shall handle the case where there is no such
+    /// rule present gracefully.
+    ///
+    /// Warning: This function is *not* async cancel safe.
+    async fn revoke_jump(
+        &self,
+        conf: &Config,
+        remote_addr: IpAddr,
+        targets: &FirewallJumpTargets,
+    ) -> ah::Result<()>;
 }
 
 // vim: ts=4 sw=4 expandtab
