@@ -1,6 +1,6 @@
 // -*- coding: utf-8 -*-
 //
-// Copyright (C) 2024 Michael Büsch <m@bues.ch>
+// Copyright (C) 2024 - 2026 Michael Büsch <m@bues.ch>
 //
 // Licensed under the Apache License version 2.0
 // or the MIT license, at your option.
@@ -208,8 +208,8 @@ async fn async_main(opts: Arc<Opts>) -> ah::Result<()> {
     let mut sighup = signal(SignalKind::hangup()).unwrap();
 
     // Create async IPC channels.
-    let (exit_sock_tx, mut exit_sock_rx) = sync::mpsc::channel(1);
-    let (exit_fw_tx, mut exit_fw_rx) = sync::mpsc::channel(1);
+    let (exit_tx, mut exit_rx) = sync::mpsc::channel(1);
+    let exit_tx = Arc::new(exit_tx);
 
     // Start the firewall unix domain socket listener.
     let srv = FirewallServer::new(opts.no_systemd, &opts)
@@ -228,6 +228,7 @@ async fn async_main(opts: Arc<Opts>) -> ah::Result<()> {
         let conf = Arc::clone(&conf);
         let opts = Arc::clone(&opts);
         let fw = Arc::clone(&fw);
+        let exit_tx = Arc::clone(&exit_tx);
 
         async move {
             let conn_semaphore = Arc::new(Semaphore::new(opts.num_connections));
@@ -248,7 +249,7 @@ async fn async_main(opts: Arc<Opts>) -> ah::Result<()> {
                         }
                     }
                     Err(e) => {
-                        let _ = exit_sock_tx.send(Err(e)).await;
+                        let _ = exit_tx.send(Err(e)).await;
                         break;
                     }
                 }
@@ -260,13 +261,14 @@ async fn async_main(opts: Arc<Opts>) -> ah::Result<()> {
     task::spawn({
         let conf = Arc::clone(&conf);
         let fw = Arc::clone(&fw);
+        let exit_tx = Arc::clone(&exit_tx);
 
         async move {
             let mut interval = time::interval(FW_MAINTAIN_PERIOD);
             loop {
                 interval.tick().await;
                 if let Err(e) = fw.maintain(&conf).await {
-                    let _ = exit_fw_tx.send(Err(e)).await;
+                    let _ = exit_tx.send(Err(e)).await;
                     break;
                 }
             }
@@ -274,31 +276,25 @@ async fn async_main(opts: Arc<Opts>) -> ah::Result<()> {
     });
 
     // Task: Main loop.
-    let mut exitcode;
-    loop {
+    let mut exitcode = loop {
         tokio::select! {
-            _ = sigterm.recv() => {
-                eprintln!("SIGTERM: Terminating.");
-                exitcode = Ok(());
-                break;
+            biased;
+            code = exit_rx.recv() => {
+                break code.unwrap_or_else(|| Err(err!("Unknown error code.")));
             }
             _ = sigint.recv() => {
-                exitcode = Err(err!("Interrupted by SIGINT."));
-                break;
+                eprintln!("Interrupted by SIGINT.");
+                break Ok(());
+            }
+            _ = sigterm.recv() => {
+                eprintln!("SIGTERM: Terminating.");
+                break Ok(());
             }
             _ = sighup.recv() => {
                 eprintln!("SIGHUP: Reloading is not supported. Please restart letmeinfwd instead.");
             }
-            code = exit_sock_rx.recv() => {
-                exitcode = code.unwrap_or_else(|| Err(err!("Unknown error code.")));
-                break;
-            }
-            code = exit_fw_rx.recv() => {
-                exitcode = code.unwrap_or_else(|| Err(err!("Unknown error code.")));
-                break;
-            }
         }
-    }
+    };
 
     // Exiting...
     // Try to remove all firewall rules.
